# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings


def show_warning(message: str, stacklevel: int = 1) -> None:
    """
    This is a helper method for within the library to ensure warnings are displayed. Jupyter notebooks
    within VSCode suppress warnings by default, so this way ensures that they are shown.

    NOTE: This is a private method only meant for use within the library.
    """
    warnings.simplefilter("always", UserWarning)
    warnings.warn(message, UserWarning, stacklevel=stacklevel)
    warnings.resetwarnings()



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import requests
import numpy as np
from .classes_forward import ATMForwardResult, ATMForwardResultRequestInfo, ATMForwardRequest
from ...exceptions import SRSAPIError


def forward(srs_obj, timestamp, geodetic_latitude, geodetic_longitude, output, maxwellian_energy_flux, gaussian_energy_flux,
            maxwellian_characteristic_energy, gaussian_peak_energy, gaussian_spectral_width, nrlmsis_model_version, oxygen_correction_factor,
            timescale_auroral, timescale_transport, atm_model_version, custom_spectrum, no_cache, timeout):
    # set timeout
    if (timeout is None):
        timeout = srs_obj.api_timeout

    # cast into request object
    request_obj = ATMForwardRequest(
        atm_model_version=atm_model_version,
        timestamp=timestamp,
        geodetic_latitude=geodetic_latitude,
        geodetic_longitude=geodetic_longitude,
        maxwellian_energy_flux=maxwellian_energy_flux,
        gaussian_energy_flux=gaussian_energy_flux,
        maxwellian_characteristic_energy=maxwellian_characteristic_energy,
        gaussian_peak_energy=gaussian_peak_energy,
        gaussian_spectral_width=gaussian_spectral_width,
        nrlmsis_model_version=nrlmsis_model_version,
        oxygen_correction_factor=oxygen_correction_factor,
        timescale_auroral=timescale_auroral,
        timescale_transport=timescale_transport,
        output=output,
        no_cache=no_cache,
        custom_spectrum=custom_spectrum,
    )

    # set up request
    post_data = {
        "atm_model_version": atm_model_version,
        "timestamp": timestamp.strftime("%Y-%m-%dT%H:%M:%S"),
        "geodetic_latitude": geodetic_latitude,
        "geodetic_longitude": geodetic_longitude,
        "maxwellian_energy_flux": maxwellian_energy_flux,
        "gaussian_energy_flux": gaussian_energy_flux,
        "maxwellian_characteristic_energy": maxwellian_characteristic_energy,
        "gaussian_peak_energy": gaussian_peak_energy,
        "gaussian_spectral_width": gaussian_spectral_width,
        "nrlmsis_model_version": nrlmsis_model_version,
        "oxygen_correction_factor": oxygen_correction_factor,
        "timescale_auroral": timescale_auroral,
        "timescale_transport": timescale_transport,
        "output": output.__dict__,
        "no_cache": no_cache,
    }

    # inject custom spectrum if supplied
    if (custom_spectrum is not None):
        post_data["custom_spectrum"] = {
            "energy": custom_spectrum[:, 0].tolist(),
            "flux": custom_spectrum[:, 1].tolist(),
        }

    # make request
    try:
        url = "%s/api/v1/atm/forward" % (srs_obj.api_base_url)
        r = requests.post(url, json=post_data, headers=srs_obj.api_headers, timeout=timeout)
    except Exception as e:  # pragma: nocover
        raise SRSAPIError("Unexpected API error: %s" % (str(e))) from e
    if (r.status_code != 200):  # pragma: nocover
        try:
            res = r.json()
            msg = res["detail"]
        except Exception:
            msg = r.content
        raise SRSAPIError("API error code %d: %s" % (r.status_code, msg))
    res = r.json()

    # cast result into return object
    request_info_obj = ATMForwardResultRequestInfo(request=request_obj, calculation_duration_ms=res["calculation_duration_ms"])
    result_obj = ATMForwardResult(
        request_info=request_info_obj,
        height_integrated_rayleighs_4278=res["data"]["height_integrated_rayleighs_4278"],
        height_integrated_rayleighs_5577=res["data"]["height_integrated_rayleighs_5577"],
        height_integrated_rayleighs_6300=res["data"]["height_integrated_rayleighs_6300"],
        height_integrated_rayleighs_8446=res["data"]["height_integrated_rayleighs_8446"],
        height_integrated_rayleighs_lbh=res["data"]["height_integrated_rayleighs_lbh"],
        height_integrated_rayleighs_1304=res["data"]["height_integrated_rayleighs_1304"],
        height_integrated_rayleighs_1356=res["data"]["height_integrated_rayleighs_1356"],
        altitudes=np.asarray(res["data"]["altitudes"]) if res["data"]["altitudes"] is not None else None,
        emission_4278=np.asarray(res["data"]["emission_4278"]) if res["data"]["emission_4278"] is not None else None,
        emission_5577=np.asarray(res["data"]["emission_5577"]) if res["data"]["emission_5577"] is not None else None,
        emission_6300=np.asarray(res["data"]["emission_6300"]) if res["data"]["emission_6300"] is not None else None,
        emission_8446=np.asarray(res["data"]["emission_8446"]) if res["data"]["emission_8446"] is not None else None,
        emission_lbh=np.asarray(res["data"]["emission_lbh"]) if res["data"]["emission_lbh"] is not None else None,
        emission_1304=np.asarray(res["data"]["emission_1304"]) if res["data"]["emission_1304"] is not None else None,
        emission_1356=np.asarray(res["data"]["emission_1356"]) if res["data"]["emission_1356"] is not None else None,
        plasma_electron_density=np.asarray(res["data"]["plasma_electron_density"]) if res["data"]["plasma_electron_density"] is not None else None,
        plasma_o2plus_density=np.asarray(res["data"]["plasma_o2plus_density"]) if res["data"]["plasma_o2plus_density"] is not None else None,
        plasma_noplus_density=np.asarray(res["data"]["plasma_noplus_density"]) if res["data"]["plasma_noplus_density"] is not None else None,
        plasma_oplus_density=np.asarray(res["data"]["plasma_oplus_density"]) if res["data"]["plasma_oplus_density"] is not None else None,
        plasma_ionisation_rate=np.asarray(res["data"]["plasma_ionisation_rate"]) if res["data"]["plasma_ionisation_rate"] is not None else None,
        plasma_electron_temperature=np.asarray(res["data"]["plasma_electron_temperature"])
        if res["data"]["plasma_electron_temperature"] is not None else None,
        plasma_ion_temperature=np.asarray(res["data"]["plasma_ion_temperature"]) if res["data"]["plasma_ion_temperature"] is not None else None,
        plasma_pederson_conductivity=np.asarray(res["data"]["plasma_pederson_conductivity"])
        if res["data"]["plasma_pederson_conductivity"] is not None else None,
        plasma_hall_conductivity=np.asarray(res["data"]["plasma_hall_conductivity"]) if res["data"]["plasma_hall_conductivity"] is not None else None,
        neutral_o2_density=np.asarray(res["data"]["neutral_o2_density"]) if res["data"]["neutral_o2_density"] is not None else None,
        neutral_o_density=np.asarray(res["data"]["neutral_o_density"]) if res["data"]["neutral_o_density"] is not None else None,
        neutral_n2_density=np.asarray(res["data"]["neutral_n2_density"]) if res["data"]["neutral_n2_density"] is not None else None,
        neutral_n_density=np.asarray(res["data"]["neutral_n_density"]) if res["data"]["neutral_n_density"] is not None else None,
        neutral_temperature=np.asarray(res["data"]["neutral_temperature"]) if res["data"]["neutral_temperature"] is not None else None,
    )

    # return
    return result_obj



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Classes for representing ATM forward calculation requests and responses. All
classes in this module are included at the top level of this library.
"""

import datetime
from dataclasses import dataclass
from typing import Optional, Literal, Any
from numpy import ndarray


@dataclass
class ATMForwardOutputFlags:
    """
    Class to represent all output values included in an ATM forward calculation.
    ATM calculations are performed in a way where you can toggle ON/OFF whichever
    pieces of information you do or don't want. This improves efficiency of the 
    calculation routine resulting in faster queries.

    By default, all output flags are set to False. There exist several helper methods
    to toggle all to True, toggle all to False, or toggle only height-integrated Rayleighs 
    flags to True. See the below Methods section for details.

    Details about each output value can be found in the documentation for the 
    [`ATMForwardResult`](classes_forward.html#pyucalgarysrs.models.atm.classes_forward.ATMForwardResult)
    object.
    """
    height_integrated_rayleighs_4278: bool = False
    height_integrated_rayleighs_5577: bool = False
    height_integrated_rayleighs_6300: bool = False
    height_integrated_rayleighs_8446: bool = False
    height_integrated_rayleighs_lbh: bool = False
    height_integrated_rayleighs_1304: bool = False
    height_integrated_rayleighs_1356: bool = False
    altitudes: bool = False
    emission_4278: bool = False
    emission_5577: bool = False
    emission_6300: bool = False
    emission_8446: bool = False
    emission_lbh: bool = False
    emission_1304: bool = False
    emission_1356: bool = False
    plasma_electron_density: bool = False
    plasma_o2plus_density: bool = False
    plasma_oplus_density: bool = False
    plasma_noplus_density: bool = False
    plasma_ionisation_rate: bool = False
    plasma_electron_temperature: bool = False
    plasma_ion_temperature: bool = False
    plasma_pederson_conductivity: bool = False
    plasma_hall_conductivity: bool = False
    neutral_o2_density: bool = False
    neutral_o_density: bool = False
    neutral_n2_density: bool = False
    neutral_n_density: bool = False
    neutral_temperature: bool = False

    def set_all_true(self):
        """
        Sets all flags to `True`.
        """
        excluded_items = ["set_all_true", "set_all_false", "enable_only_height_integrated_rayleighs"]
        for var_name in dir(self):
            if (not var_name.startswith("__") and var_name not in excluded_items):
                setattr(self, var_name, True)

    def set_all_false(self):
        """
        Sets all flags to `False`.
        """
        excluded_items = ["set_all_true", "set_all_false", "enable_only_height_integrated_rayleighs"]
        for var_name in dir(self):
            if (not var_name.startswith("__") and var_name not in excluded_items):
                setattr(self, var_name, False)

    def enable_only_height_integrated_rayleighs(self):
        """
        Sets only height-integrated Rayleighs values to `True`.
        """
        for var_name in filter(lambda x: x.startswith("height_integrated_rayleighs"), dir(self)):
            setattr(self, var_name, True)


@dataclass
class ATMForwardRequest:
    """
    Class that represents the UCalgary Space Remote Sensing API request when 
    performing an ATM forward calculation. This object is included in any
    [`ATMForwardResultRequestInfo`](classes_forward.html#pyucalgarysrs.models.atm.classes_forward.ATMForwardResultRequestInfo) 
    and [`ATMForwardResult`](classes_forward.html#pyucalgarysrs.models.atm.classes_forward.ATMForwardResult) 
    objects.
    """
    atm_model_version: Literal["1.0"]
    timestamp: datetime.datetime
    geodetic_latitude: float
    geodetic_longitude: float
    maxwellian_energy_flux: float
    gaussian_energy_flux: float
    maxwellian_characteristic_energy: float
    gaussian_peak_energy: float
    gaussian_spectral_width: float
    nrlmsis_model_version: Literal["00", "2.0"]
    oxygen_correction_factor: float
    timescale_auroral: int
    timescale_transport: int
    custom_spectrum: Optional[ndarray]
    output: ATMForwardOutputFlags
    no_cache: bool


@dataclass
class ATMForwardResultRequestInfo:
    """
    Class containing details about interacting with the UCalgary Space Remote Sensing
    API when performing an ATM forward calculation. It contains the API request itself,
    and other information returned by the API.

    An instance of this is included in any
    [`ATMForwardResult`](classes_forward.html#pyucalgarysrs.models.atm.classes_forward.ATMForwardResult) 
    object.

    Attributes:
        request (ATMForwardRequest): 
            Instance of class that represents the lower-level API request when performing 
            an ATM forward calculation.

        calculation_duration_ms (float): 
            Duration the the API spent performing the ATM forward calculation. Represented
            in milliseconds.
    """
    request: ATMForwardRequest
    calculation_duration_ms: float


@dataclass
class ATMForwardResult:
    """
    Class containing all data from an ATM forward calculation. This class also includes
    details about the input parameters to the forward calculation routine (`request_info`),
    along with all output data values.
    
    Based on the request's output flags, if flags were set to False then the values will 
    be None types. If flags were set to True then the values will be their respective 
    types (e.g., float, numpy ndarray).

    Attributes:
        request_info (ATMForwardResultRequestInfo): 
            Information about the API request made to perform the ATM forward calculation.
        
        height_integrated_rayleighs_4278 (float): 
            Height-integrated Rayleighs value for the 427.8nm emission (blue).

        height_integrated_rayleighs_5577 (float): 
            Height-integrated Rayleighs value for the 557.7nm emission (green).
            
        height_integrated_rayleighs_6300 (float): 
            Height-integrated Rayleighs value for the 630.0nm emission(red).
        
        height_integrated_rayleighs_8446 (float): 
            Height-integrated Rayleighs value for the 844.6nm emission (near infrared).

        height_integrated_rayleighs_lbh (float): 
            Height-integrated Rayleighs value for the Lyman-Birge-Hopfield emission.

        height_integrated_rayleighs_1304 (float): 
            Height-integrated Rayleighs value for the 130.4nm emission.

        height_integrated_rayleighs_1356 (float): 
            Height-integrated Rayleighs value for the 135.6nm emission.

        altitudes (ndarray): 
            A 1-dimensional numpy array for the altitudes in kilometers.

        emission_4278 (ndarray): 
            A 1-dimensional numpy array for the 427.8nm volume emission rate (1/cm^3/s).

        emission_5577 (ndarray): 
            A 1-dimensional numpy array for the 557.7nm volume emission rate (1/cm^3/s).

        emission_6300 (ndarray): 
            A 1-dimensional numpy array for the 630.0nm volume emission rate (1/cm^3/s).

        emission_8446 (ndarray): 
            A 1-dimensional numpy array for the 844.6nm volume emission rate (1/cm^3/s).

        emission_lbh (ndarray): 
            A 1-dimensional numpy array for the Lyman-Birge-Hopfield volume emission rate (1/cm^3/s).

        emission_1304 (ndarray): 
            A 1-dimensional numpy array for the 130.4nm volume emission rate (1/cm^3/s).

        emission_1356 (ndarray): 
            A 1-dimensional numpy array for the 135.6nm volume emission rate (1/cm^3/s).

        plasma_electron_density (ndarray): 
            A 1-dimensional numpy array for the plasma electron density (cm^-3).

        plasma_o2plus_density (ndarray): 
            A 1-dimensional numpy array for the plasma O2+ density (cm^-3).

        plasma_noplus_density (ndarray): 
            A 1-dimensional numpy array for the plasma NO+ density (cm^-3).

        plasma_oplus_density (ndarray): 
            A 1-dimensional numpy array for the plasma O+ density (cm^-3).

        plasma_ionisation_rate (ndarray): 
            A 1-dimensional numpy array for the plasma ionisation rate (1/cm^3/s).

        plasma_electron_temperature (ndarray): 
            A 1-dimensional numpy array for the plasma electron temperature (Kelvin).

        plasma_ion_temperature (ndarray): 
            A 1-dimensional numpy array for the plasma ion temperature (Kelvin).

        plasma_pederson_conductivity (ndarray): 
            A 1-dimensional numpy array for the Peterson plasma conductivity (S/m).

        plasma_hall_conductivity (ndarray): 
            A 1-dimensional numpy array for the hall plasma conductivity (S/m).

        neutral_o2_density (ndarray): 
            A 1-dimensional numpy array for the neutral O2 density (cm^-3).

        neutral_o_density (ndarray): 
            A 1-dimensional numpy array for the neutral O density (cm^-3).

        neutral_n2_density (ndarray): 
            A 1-dimensional numpy array for the neutral N2 density (cm^-3).

        neutral_n_density (ndarray): 
            A 1-dimensional numpy array for the neutral N density (cm^-3).

        neutral_temperature (ndarray): 
            A 1-dimensional numpy array for the neutral temperature (Kelvin).
    """
    request_info: ATMForwardResultRequestInfo
    height_integrated_rayleighs_4278: Any
    height_integrated_rayleighs_5577: Any
    height_integrated_rayleighs_6300: Any
    height_integrated_rayleighs_8446: Any
    height_integrated_rayleighs_lbh: Any
    height_integrated_rayleighs_1304: Any
    height_integrated_rayleighs_1356: Any
    altitudes: Any
    emission_4278: Any
    emission_5577: Any
    emission_6300: Any
    emission_8446: Any
    emission_lbh: Any
    emission_1304: Any
    emission_1356: Any
    plasma_electron_density: Any
    plasma_o2plus_density: Any
    plasma_noplus_density: Any
    plasma_oplus_density: Any
    plasma_ionisation_rate: Any
    plasma_electron_temperature: Any
    plasma_ion_temperature: Any
    plasma_pederson_conductivity: Any
    plasma_hall_conductivity: Any
    neutral_o2_density: Any
    neutral_o_density: Any
    neutral_n2_density: Any
    neutral_n_density: Any
    neutral_temperature: Any

    def pretty_print(self):
        """
        A special print output for this class.
        """
        print("ATMForwardResult:")
        for var_name in dir(self):
            # exclude methods
            if (var_name.startswith("__") or var_name == "pretty_print"):
                continue

            # convert var to string format we want
            var_value = getattr(self, var_name)
            var_str = "None"
            if (var_name == "request_info"):
                var_str = "ATMForwardResultRequestInfo(...)"
            if (var_value is not None):
                if (isinstance(var_value, float)):
                    var_str = "%f" % (var_value)
                elif (isinstance(var_value, ndarray)):
                    var_str = "%s ...])" % (var_value.__repr__()[0:60])

            # print string for this var
            print("  %-37s: %s" % (var_name, var_str))



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import signal
import numpy as np
import h5py
import os
from pathlib import Path
from multiprocessing import Pool
from functools import partial
from ..classes import HSRData

# globals
HSR_DT = np.dtype("float32")


def read(file_list, n_parallel=1, no_metadata=False, start_time=None, end_time=None, quiet=False):
    # if input is just a single file name in a string, convert to a list to be fed to the workers
    if isinstance(file_list, str) or isinstance(file_list, Path):
        file_list = [file_list]

    # check n_parallel
    if (n_parallel > 1):
        try:
            # set up process pool (ignore SIGINT before spawning pool so child processes inherit SIGINT handler)
            original_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
            pool = Pool(processes=n_parallel)
            signal.signal(signal.SIGINT, original_sigint_handler)  # restore SIGINT handler
        except ValueError:  # pragma: nocover
            # likely the read call is being used within a context that doesn't support the usage
            # of signals in this way, proceed without it
            pool = Pool(processes=n_parallel)

        # call readfile function, run each iteration with a single input file from file_list
        # NOTE: structure of data - data[file][metadata dictionary lists = 1, images = 0][frame]
        data = []
        try:
            data = pool.map(partial(
                __hsr_readfile_worker,
                no_metadata=no_metadata,
                start_time=start_time,
                end_time=end_time,
                quiet=quiet,
            ), file_list)
        except KeyboardInterrupt:  # pragma: nocover
            pool.terminate()  # gracefully kill children
            return [], [], [], []
        else:
            pool.close()
            pool.join()
    else:
        # don't bother using multiprocessing with one worker, just call the worker function directly
        data = []
        for f in file_list:
            data.append(__hsr_readfile_worker(
                f,
                no_metadata=no_metadata,
                start_time=start_time,
                end_time=end_time,
                quiet=quiet,
            ))

    # compile results
    hsr_data = []
    top_level_timestamps = []
    metadata = []
    problematic_file_list = []
    for result in data:
        # problematic
        if (result["problematic"] is True):
            problematic_file_list.append({
                "filename": result["file"],
                "error_message": result["error_message"],
            })
            continue

        # skip empty metadata
        if (result["metadata"] is None):
            continue

        # hsr data
        hsr_data.append(
            HSRData(
                band_central_frequency=result["band_central_frequency_list"],
                band_passband=result["band_passband_list"],
                timestamp=result["np_timestamp"],
                raw_power=result["np_raw_power"],
                absorption=result["np_absorption"],
            ))

        # timestamps
        if (len(result["np_timestamp"]) > 0):
            top_level_timestamps.append(result["np_timestamp"][0])

        # metadata
        if (no_metadata is False):
            metadata.append(result["metadata"])

    # return
    return hsr_data, top_level_timestamps, metadata, problematic_file_list


def __str_to_datetime_formatter(timestamp_str):
    return datetime.datetime.strptime(timestamp_str.decode(), "%Y-%m-%d %H:%M:%S UTC")


def __hsr_readfile_worker(file, no_metadata=False, start_time=None, end_time=None, quiet=False):
    # init
    metadata_dict = {}
    band_central_frequency_list = []
    band_passband_list = []
    np_timestamp = np.array([], dtype=datetime.datetime)
    np_raw_power = np.array([], dtype=HSR_DT)
    np_absorption = np.array([], dtype=HSR_DT)
    problematic = False
    error_message = ""

    # extract start and end times of the filename
    try:
        file_dt = datetime.datetime.strptime(os.path.basename(file)[0:8], "%Y%m%d")
    except Exception:
        if (quiet is False):
            print("Failed to extract timestamp from filename")
        problematic = True
        error_message = "failed to extract timestamp from filename"
        return {
            "band_central_frequency_list": band_central_frequency_list,
            "band_passband_list": band_passband_list,
            "np_timestamp": np_timestamp,
            "np_raw_power": np_raw_power,
            "np_absorption": np_absorption,
            "metadata": metadata_dict,
            "problematic": problematic,
            "file": file,
            "error_message": error_message,
        }

    # cross-check the filename with the start and end times; this will allow
    # files that are outside of the desired time frame to not actually bother
    # with getting read
    if ((start_time is None or file_dt >= start_time.replace(hour=0, minute=0, second=0, microsecond=0))
            and (end_time is None or file_dt <= end_time.replace(hour=0, minute=0, second=0, microsecond=0))):
        # this file should be read
        pass
    else:
        # this file doesn't need to be read
        return {
            "band_central_frequency_list": band_central_frequency_list,
            "band_passband_list": band_passband_list,
            "np_timestamp": np_timestamp,
            "np_raw_power": np_raw_power,
            "np_absorption": np_absorption,
            "metadata": None,  # the key that triggers this file to be skipped at post-processing
            "problematic": problematic,
            "file": file,
            "error_message": error_message,
        }

    # process file
    try:
        # open H5 file
        f = h5py.File(file, 'r')

        # get timestamp, convert into datetime objects
        np_timestamp = f["data"]["timestamp"][:]  # type: ignore
        np_timestamp = np.vectorize(__str_to_datetime_formatter)(np_timestamp)  # type: ignore
        np_timestamp = np_timestamp.astype(datetime.datetime)

        # determine indexes to keep based on start and end times
        idxs = []
        for idx, val in enumerate(np_timestamp):
            if ((start_time is None or val >= start_time) and (end_time is None or val <= end_time)):
                idxs.append(idx)
        np_timestamp = np_timestamp[idxs]

        # get raw power, timestamp
        np_raw_power = f["data"]["raw_power"][:, idxs]  # type: ignore

        # get band central frequency, bandpass
        band_central_frequency_list = f["data"]["band_central_frequency"][:].tolist()  # type: ignore
        band_central_frequency_list = [x.decode() for x in band_central_frequency_list]
        band_passband_list = f["data"]["band_passband"][:].tolist()  # type: ignore
        band_passband_list = [x.decode() for x in band_passband_list]

        # set absorption
        if ("_k0_" in os.path.basename(file)):
            # k0 data, no absorption
            np_absorption = None

        # get metadata
        if (no_metadata is False):
            for key, value in f["metadata"]["file"].attrs.items():  # type: ignore
                metadata_dict[key] = value

        # close file
        f.close()
    except Exception as e:
        # set error message
        if (quiet is False):
            print("Failed to read file '%s': %s" % (file, str(e)))
        problematic = True
        error_message = "failed to open file: %s" % (str(e))

        # clear values
        band_central_frequency_list = []
        band_passband_list = []
        np_timestamp = np.array([], dtype=datetime.datetime)
        np_raw_power = np.array([], dtype=HSR_DT)
        np_absorption = np.array([], dtype=HSR_DT)

        # close file
        try:
            f.close()  # type: ignore
        except Exception:
            pass

        # return
        return {
            "band_central_frequency_list": band_central_frequency_list,
            "band_passband_list": band_passband_list,
            "np_timestamp": np_timestamp,
            "np_raw_power": np_raw_power,
            "np_absorption": np_absorption,
            "metadata": metadata_dict,
            "problematic": problematic,
            "file": file,
            "error_message": error_message,
        }

    # return
    return {
        "band_central_frequency_list": band_central_frequency_list,
        "band_passband_list": band_passband_list,
        "np_timestamp": np_timestamp,
        "np_raw_power": np_raw_power,
        "np_absorption": np_absorption,
        "metadata": metadata_dict,
        "problematic": problematic,
        "file": file,
        "error_message": error_message,
    }



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import requests
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from .classes import FileListingResponse, FileDownloadResult, Dataset
from ..exceptions import SRSAPIError, SRSDownloadError
from .._util import show_warning


def __download_url(
    url,
    prefix,
    output_base_path,
    headers,
    timeout,
    overwrite=False,
    pbar=None,
    pbar_iterator_nfiles=False,
):
    # set output filename
    output_filename = Path(output_base_path) / Path(url.removeprefix(prefix + "/"))
    if (overwrite is False and os.path.exists(output_filename)):
        if (pbar is not None):
            if (pbar_iterator_nfiles is True):
                pbar.update()
            else:
                pbar.update(os.path.getsize(output_filename))
        return {"filename": output_filename, "bytes_downloaded": 0}

    # create destination directory
    try:
        os.makedirs(os.path.dirname(output_filename), exist_ok=True)
    except Exception:  # pragma: nocover
        # NOTE: sometimes when making directories in parallel there are race conditions. We put
        # in a catch here and carry on if there are ever issues.
        pass

    # retrieve file and save to disk
    r = requests.get(url, headers=headers, timeout=timeout)
    if (r.status_code == 200):
        this_bytes = len(r.content)
        with open(output_filename, 'wb') as fp:
            fp.write(r.content)

        # advance progress
        if (pbar is not None):
            if (pbar_iterator_nfiles is True):
                pbar.update()
            else:
                pbar.update(this_bytes)
    else:
        if (pbar is not None):
            if (pbar_iterator_nfiles is True):
                pbar.update()
            else:
                pbar.update(0)  # pragma: no cover
        raise SRSDownloadError("HTTP error %d when downloading '%s'" % (r.status_code, url))

    # return filename
    return {"filename": output_filename, "bytes_downloaded": this_bytes}


def __download_urls(srs_obj,
                    file_listing_obj,
                    n_parallel,
                    overwrite,
                    progress_bar_disable,
                    progress_bar_ncols,
                    progress_bar_ascii,
                    progress_bar_desc,
                    timeout,
                    progress_bar_format_numurls_nobytes=False):

    def __do_parallel_work(pbar=None, pbar_iterator_nfiles=False):

        def download_task(i):
            return __download_url(
                file_listing_obj.urls[i],
                path_prefix,
                output_path,
                srs_obj.api_headers,
                timeout,
                overwrite=overwrite,
                pbar=pbar,
                pbar_iterator_nfiles=pbar_iterator_nfiles,
            )

        with ThreadPoolExecutor(max_workers=n_parallel) as executor:
            futures = [executor.submit(download_task, i) for i in range(len(file_listing_obj.urls))]
            job_data = [future.result() for future in futures]

        return job_data

    # set output path
    output_path = srs_obj.download_output_root_path / Path(file_listing_obj.dataset.name)

    # check if there's files to download
    if (file_listing_obj.count == 0):
        show_warning("No data found to download", stacklevel=5)

        # cast into return obj
        download_obj = FileDownloadResult(
            filenames=[],
            count=0,
            dataset=file_listing_obj.dataset,
            total_bytes=0,
            output_root_path=output_path,
        )
    else:
        # set timeout
        if (timeout is None):
            timeout = srs_obj.api_timeout

        # set path prefix
        path_prefix = file_listing_obj.path_prefix

        # set progress bar description text
        desc_str = "Downloading %s files" % (file_listing_obj.dataset.name)
        if (progress_bar_desc is not None):
            desc_str = progress_bar_desc

        # download the urls
        parallel_data = []
        if (progress_bar_disable is True):
            parallel_data = __do_parallel_work()
        else:
            if (progress_bar_format_numurls_nobytes is True):
                # total bytes could be inaccurate, progress bar should use count of urls
                # and iterator of 'files' instead of bytes.
                with srs_obj._tqdm(total=len(file_listing_obj.urls), desc=desc_str, unit="files", ncols=progress_bar_ncols,
                                   ascii=progress_bar_ascii) as pbar:
                    parallel_data = __do_parallel_work(pbar=pbar, pbar_iterator_nfiles=True)
            else:
                # total bytes is accurate, show MB/s
                with srs_obj._tqdm(total=file_listing_obj.total_bytes,
                                   desc=desc_str,
                                   ncols=progress_bar_ncols,
                                   ascii=progress_bar_ascii,
                                   unit="B",
                                   unit_scale=True) as pbar:
                    parallel_data = __do_parallel_work(pbar=pbar)

        # cast into return obj
        filenames_list = []
        total_bytes_downloaded = 0
        for p in parallel_data:
            filenames_list.append(p["filename"])  # type: ignore
            total_bytes_downloaded += p["bytes_downloaded"]  # type: ignore
        download_obj = FileDownloadResult(
            filenames=filenames_list,
            count=len(parallel_data),
            dataset=file_listing_obj.dataset,
            total_bytes=total_bytes_downloaded,
            output_root_path=output_path,
        )

    # return
    return download_obj


def get_urls(srs_obj, dataset_name, start, end, site_uid, device_uid, timeout, warning_stack_level=3):
    # set timeout
    if (timeout is None):
        timeout = srs_obj.api_timeout

    # set up API file listing request
    params = {
        "name": dataset_name,
        "start": start,
        "end": end,
        "include_total_bytes": True,
    }
    if (site_uid is not None):
        params["site_uid"] = site_uid
    if (device_uid is not None):
        params["device_uid"] = device_uid

    # check warnings about site and device being supplied
    if (site_uid is not None and "calibration" in dataset_name.lower()):
        show_warning("The site_uid filter was used when retrieving URLs for the dataset %s. Please note "
                     "that the site_uid field is not used when filtering in this dataset. Consider removing "
                     "it from your function call." % (dataset_name),
                     stacklevel=warning_stack_level)
    if (device_uid is not None and "skymap" in dataset_name.lower()):
        show_warning("The device_uid filter was used when retrieving URLs for the dataset %s. Please note "
                     "that the device_uid field is not used when filtering in this dataset. Consider removing "
                     "it from your function call." % (dataset_name),
                     stacklevel=warning_stack_level)

    # make API request
    url = "%s/api/v1/data_distribution/urls" % (srs_obj.api_base_url)
    try:
        r = requests.get(url, params=params, headers=srs_obj.api_headers, timeout=timeout)
    except Exception as e:  # pragma: nocover
        raise SRSAPIError("Unexpected API error: %s" % (str(e))) from e
    if (r.status_code != 200):  # pragma: nocover
        try:
            res = r.json()
            msg = res["detail"]
        except Exception:
            msg = r.content
        raise SRSAPIError("API error code %d: %s" % (r.status_code, msg))
    res = r.json()

    # get list of file reading supported datasets
    file_reading_supported_datasets = srs_obj.data.list_supported_read_datasets()

    # cast response into FileDownloadResult object
    file_listing_obj = FileListingResponse(**res)

    # cast dataset part of response
    file_reading_supported = True if res["dataset"]["name"] in file_reading_supported_datasets else False
    file_listing_obj.dataset = Dataset(**res["dataset"], file_reading_supported=file_reading_supported)

    # return
    return file_listing_obj


def download_generic(srs_obj, dataset_name, start, end, site_uid, device_uid, n_parallel, overwrite, progress_bar_disable, progress_bar_ncols,
                     progress_bar_ascii, progress_bar_desc, timeout):
    # get file listing
    file_listing_obj = get_urls(srs_obj, dataset_name, start, end, site_uid, device_uid, timeout, warning_stack_level=4)

    # download the urls
    download_obj = __download_urls(
        srs_obj,
        file_listing_obj,
        n_parallel,
        overwrite,
        progress_bar_disable,
        progress_bar_ncols,
        progress_bar_ascii,
        progress_bar_desc,
        timeout,
    )

    # return
    return download_obj


def download_using_urls(srs_obj, file_listing_obj, n_parallel, overwrite, progress_bar_disable, progress_bar_ncols, progress_bar_ascii,
                        progress_bar_desc, timeout):
    # download the urls
    download_obj = __download_urls(srs_obj,
                                   file_listing_obj,
                                   n_parallel,
                                   overwrite,
                                   progress_bar_disable,
                                   progress_bar_ncols,
                                   progress_bar_ascii,
                                   progress_bar_desc,
                                   timeout,
                                   progress_bar_format_numurls_nobytes=True)

    # return
    return download_obj



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import humanize
from texttable import Texttable
from typing import Optional, Dict, Any, Literal
from pathlib import Path
from tqdm import tqdm as tqdm_std
from tqdm.auto import tqdm as tqdm_auto
from tqdm.notebook import tqdm as tqdm_notebook
from .exceptions import SRSInitializationError, SRSPurgeError
from .data import DataManager
from .models import ModelsManager
from . import __version__
from ._util import show_warning


class PyUCalgarySRS:
    """
    The `PyUCalgarySRS` class is the primary entry point for utilizing
    this library. It is used to initialize a session, capturing details
    about API connectivity, environment, and more. All submodules are 
    encapsulated within this class, so any usage of the library starts 
    with creating this object.

    ```python
    import pyucalgarysrs
    srs = pyucalgarysrs.PyUCalgarySRS()
    ```

    When working with this object, you can set configuration parameters, such 
    as the destination directory for downloaded data, or API special settings 
    (e.g., timeout, HTTP headers). These parameters can be set when instantiating 
    the object, or after instantiating using the self-contained accessible variables.
    """

    __DEFAULT_API_BASE_URL = "https://api.phys.ucalgary.ca"
    __DEFAULT_API_TIMEOUT = 30
    __DEFAULT_API_HEADERS = {
        "content-type": "application/json",
        "user-agent": "python-pyucalgarysrs/%s" % (__version__),
    }  # NOTE: these MUST be lowercase so that the decorator logic cannot be overridden
    __DEFAULT_PROGRESS_BAR_BACKEND = "auto"

    def __init__(self,
                 download_output_root_path: Optional[str] = None,
                 read_tar_temp_path: Optional[str] = None,
                 api_base_url: Optional[str] = None,
                 api_timeout: Optional[int] = None,
                 api_headers: Optional[Dict] = None,
                 api_key: Optional[str] = None,
                 progress_bar_backend: Literal["auto", "standard", "notebook"] = "auto"):
        """
        Attributes:
            download_output_root_path (str): 
                Destination directory for downloaded data. The default for this path is a 
                subfolder in the user's home directory, such  as `/home/user/pyucalgarysrs_data` 
                in Linux. In Windows and Mac, it is similar.

            read_tar_temp_path (str): 
                Temporary directory used for tar extraction phases during file reading (e.g., 
                reading TREx RGB Burst data). The default for this is `<download_output_root_path>/tar_temp_working`. 
                For faster performance when reading tar-based data, one option on Linux is 
                to set this to use RAM directly at `/dev/shm/pyucalgarysrs_tar_temp_working`.

            api_base_url (str): 
                URL prefix to use when interacting with the UCalgary Space Remote Sensing API. By
                default this is set to `https://api.phys.ucalgary.ca`. This parameter is primarily
                used by the development team to test and build new functions using the private staging
                API.

            api_timeout (int): 
                The timeout used when communicating with the UCalgary Space Remote Sensing API. This 
                value is represented in seconds, and by default is `30 seconds`.
            
            api_headers (Dict): 
                HTTP headers used when communicating with the UCalgary Space Remote Sensing API. The 
                default for this value consists of several standard headers. Any changes to this 
                parameter are in addition to the default standard headers.

            api_key (str): 
                API key to use when interacting with the UCalgary Space Remote Sensing API. The default
                value is None. Please note that an API key is currently not required for using the API,
                and this parameter is implemented purely for future-proofing. It is presently not utilized.
        
            progress_bar_backend (str): 
                The progress bar backend to use. Valid choices are 'auto', 'standard', or 'notebook'. 
                Default is 'auto'. This parameter is optional.

        Raises:
            pyucalgarysrs.exceptions.SRSInitializationError: an error was encountered during
                initialization of the paths
        """
        # initialize path parameters
        self.__download_output_root_path = download_output_root_path
        self.__read_tar_temp_path = read_tar_temp_path

        # initialize api parameters
        self.__api_base_url = api_base_url
        if (api_base_url is None):
            self.__api_base_url = self.__DEFAULT_API_BASE_URL
        self.__api_headers = api_headers
        if (api_headers is None):
            self.__api_headers = self.__DEFAULT_API_HEADERS
        self.__api_timeout = api_timeout
        if (api_timeout is None):
            self.__api_timeout = self.__DEFAULT_API_TIMEOUT
        self.__api_key = api_key

        # initialize progress bar parameters
        self.__progress_bar_backend = progress_bar_backend
        self._tqdm = None
        self.__initialize_progress_bar_backend()

        # initialize paths
        self.__initialize_paths()

        # initialize sub-modules
        self.__data = DataManager(self)
        self.__models = ModelsManager(self)

    # ------------------------------------------
    # properties for submodule managers
    # ------------------------------------------
    @property
    def data(self):
        """
        Access to the `data` submodule from within a PyUCalgarySRS object.
        """
        return self.__data

    @property
    def models(self):
        """
        Access to the `models` submodule from within a PyUCalgarySRS object.
        """
        return self.__models

    # ------------------------------------------
    # properties for configuration parameters
    # ------------------------------------------
    @property
    def api_base_url(self):
        """
        Property for the API base URL. See above for details.
        """
        return self.__api_base_url

    @api_base_url.setter
    def api_base_url(self, value: Optional[str] = None):
        if (value is None):
            self.__api_base_url = self.__DEFAULT_API_BASE_URL
        else:
            # check if http:// or https:// is in the URL
            value = value.lower()
            if (len(value) <= 8 or ("https://" not in value and "http://" not in value)):
                raise SRSInitializationError("API base URL is an invalid URL")

            # remove trailing slash if there is one
            if (value[-1] == '/'):
                value = value[0:-1]
            self.__api_base_url = value

    @property
    def api_headers(self):
        """
        Property for the API headers. See above for details.
        """
        return self.__api_headers

    @api_headers.setter
    def api_headers(self, value: Dict):
        new_headers = self.__DEFAULT_API_HEADERS
        if (value is not None):
            for k, v in value.items():
                k = k.lower()
                if (k in new_headers):
                    if (k == "user-agent" and "python-pyaurorax/" not in v):
                        show_warning("Cannot override default '%s' header" % (k), stacklevel=1)
                    else:
                        # allow pyaurorax to change the user agent
                        new_headers[k] = v
                else:
                    new_headers[k] = v
        self.__api_headers = new_headers

    @property
    def api_timeout(self):
        """
        Property for the API timeout. See above for details.
        """
        return self.__api_timeout

    @api_timeout.setter
    def api_timeout(self, value: Optional[int] = None):
        if (value is None):
            self.__api_timeout = self.__DEFAULT_API_TIMEOUT
        else:
            self.__api_timeout = value

    @property
    def api_key(self):
        """
        Property for the API key. See above for details.
        """
        return self.__api_key

    @api_key.setter
    def api_key(self, value: str):
        self.__api_key = value  # pragma: nocover

    @property
    def download_output_root_path(self):
        """
        Property for the download output root path. See above for details.
        """
        return str(self.__download_output_root_path)

    @download_output_root_path.setter
    def download_output_root_path(self, value: str):
        self.__download_output_root_path = value
        self.__initialize_paths()

    @property
    def read_tar_temp_path(self):
        """
        Property for the read tar temp path. See above for details.
        """
        return str(self.__read_tar_temp_path)

    @read_tar_temp_path.setter
    def read_tar_temp_path(self, value: str):
        self.__read_tar_temp_path = value
        self.__initialize_paths()

    @property
    def progress_bar_backend(self):
        """
        Property for the progress bar backend. See above for details.
        """
        return self.__progress_bar_backend

    @progress_bar_backend.setter
    def progress_bar_backend(self, value: Optional[str] = None):
        # set the backend string
        if (value is None):
            self.__progress_bar_backend = self.__DEFAULT_PROGRESS_BAR_BACKEND
        else:
            value = value.lower()
            if (value != "auto" and value != "standard" and value != "notebook"):
                raise SRSInitializationError("Invalid progress bar backend. Allowed values are 'auto', 'standard' or 'notebook'.")
            self.__progress_bar_backend = value

        # set the backend tqdm object
        if (value == "auto"):
            self._tqdm = tqdm_auto
        elif (value == "notebook"):
            self._tqdm = tqdm_notebook
        elif (value == "standard"):
            self._tqdm = tqdm_std

    # -----------------------------
    # special methods
    # -----------------------------
    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return ("PyUCalgarySRS(download_output_root_path='%s', read_tar_temp_path='%s', api_base_url='%s', api_headers=%s, api_timeout=%s)") % (
            self.__download_output_root_path,
            self.__read_tar_temp_path,
            self.api_base_url,
            self.api_headers,
            self.api_timeout,
        )

    # -----------------------------
    # private methods
    # -----------------------------
    def __initialize_paths(self):
        """
        Initialize the `download_output_root_path` and `read_tar_temp_path` directories.
        """
        if (self.__download_output_root_path is None):
            self.__download_output_root_path = Path("%s/pyucalgarysrs_data" % (str(Path.home())))
        if (self.__read_tar_temp_path is None):
            self.__read_tar_temp_path = Path("%s/tar_temp_working" % (self.__download_output_root_path))
        try:
            os.makedirs(self.download_output_root_path, exist_ok=True)
            os.makedirs(self.read_tar_temp_path, exist_ok=True)
        except IOError as e:  # pragma: nocover
            raise SRSInitializationError("Error during output path creation: %s" % str(e)) from e

    def __initialize_progress_bar_backend(self):
        """
        Initialize the `progress_bar_backend` parameter.
        
        The default of 'auto' works in all cases except for if the library is used within Spyder. In that 
        IDE, tqdm is not able to choose the backend automatically.

        NOTE: we are still working on figuring out a way to handle Spyder better.
        """
        self.progress_bar_backend = self.__progress_bar_backend

    # -----------------------------
    # public methods
    # -----------------------------
    def purge_download_output_root_path(self):
        """
        Delete all files in the `download_output_root_path` directory. Since the
        library downloads data to this directory, over time it can grow too large
        and the user can risk running out of space. This method is here to assist
        with easily clearing out this directory.

        Raises:
            pyucalgarysrs.exceptions.SRSPurgeError: an error was encountered during the purge operation
        """
        try:
            for item in os.listdir(self.download_output_root_path):
                item = Path(self.download_output_root_path) / item
                if (item.is_dir() is True and self.read_tar_temp_path != str(item)):
                    shutil.rmtree(item)
                elif (item.is_file() is True):
                    os.remove(item)
        except Exception as e:  # pragma: nocover
            raise SRSPurgeError("Error while purging download output root path: %s" % (str(e))) from e

    def purge_read_tar_temp_path(self):
        """
        Delete all files in the `read_tar_temp_path` directory. Since the library 
        extracts temporary data to this directory, sometime issues during reading 
        can cause this directory to contain residual files that aren't deleted during 
        the normal read routine. Though this is very rare, it is still possible. 
        Therefore, this method is here to assist with easily clearing out this 
        directory.

        Raises:
            pyucalgarysrs.exceptions.SRSPurgeError: an error was encountered during the purge operation
        """
        try:
            for item in os.listdir(self.read_tar_temp_path):
                item = Path(self.read_tar_temp_path) / item
                if (item.is_dir() is True and self.download_output_root_path != str(item)):
                    shutil.rmtree(item)
                elif (item.is_file() is True):
                    os.remove(item)
        except Exception as e:  # pragma: nocover
            raise SRSPurgeError("Error while purging read tar temp path: %s" % (str(e))) from e

    def show_data_usage(self, order: Literal["name", "size"] = "size", return_dict: bool = False) -> Any:
        """
        Print the volume of data existing in the download_output_root_path, broken down
        by dataset. Alternatively return the information in a dictionary.
        
        This can be a helpful tool for managing your disk space.

        Args:
            order (bool): 
                Order results by either `size` or `name`. Default is `size`.

            return_dict (bool): 
                Instead of printing the data usage information, return information as a dictionary.

        Returns:
            Printed output. If `return_dict` is True, then it will instead return a dictionary with the
            disk usage information.
        
        Notes:
            Note that size on disk may differ slightly from the values determined by this 
            routine. For example, the results here will be slightly different than the output
            of a 'du' command on *nix systems.
        """
        # init
        total_size = 0
        download_pathlib_path = Path(self.download_output_root_path)

        # get list of dataset paths
        dataset_paths = []
        for f in os.listdir(download_pathlib_path):
            path_f = download_pathlib_path / f
            if (os.path.isdir(path_f) is True and str(path_f) != self.read_tar_temp_path):
                dataset_paths.append(path_f)

        # get size of each dataset path
        dataset_dict = {}
        longest_path_len = 0
        for dataset_path in dataset_paths:
            # get size
            dataset_size = 0
            for dirpath, _, filenames in os.walk(dataset_path):
                for filename in filenames:
                    filepath = os.path.join(dirpath, filename)
                    if (os.path.isfile(filepath) is True):
                        dataset_size += os.path.getsize(filepath)

            # check if this is the longest path name
            path_basename = os.path.basename(dataset_path)
            if (longest_path_len == 0):
                longest_path_len = len(path_basename)
            elif (len(path_basename) > longest_path_len):
                longest_path_len = len(path_basename)

            # set dict
            dataset_dict[path_basename] = {
                "path_obj": dataset_path,
                "size_bytes": dataset_size,
                "size_str": humanize.naturalsize(dataset_size),
            }

            # add to total
            total_size += dataset_size

        # return dictionary
        if (return_dict is True):
            return dataset_dict

        # print table
        #
        # order into list
        order_key = "size_bytes" if order == "size" else order
        ordered_list = []
        for path, p_dict in dataset_dict.items():
            this_dict = p_dict
            this_dict["name"] = path
            ordered_list.append(this_dict)
        if (order == "size"):
            ordered_list = reversed(sorted(ordered_list, key=lambda x: x[order_key]))
        else:
            ordered_list = sorted(ordered_list, key=lambda x: x[order_key])

        # set column data
        table_names = []
        table_sizes = []
        for item in ordered_list:
            table_names.append(item["name"])
            table_sizes.append(item["size_str"])

        # set header values
        table_headers = ["Dataset name", "Size"]

        # print as table
        table = Texttable()
        table.set_deco(Texttable.HEADER)
        table.set_cols_dtype(["t"] * len(table_headers))
        table.set_header_align(["l"] * len(table_headers))
        table.set_cols_align(["l"] * len(table_headers))
        table.header(table_headers)
        for i in range(0, len(table_names)):
            table.add_row([table_names[i], table_sizes[i]])
        print(table.draw())

        print("\nTotal size: %s" % (humanize.naturalsize(total_size)))
